import sympy as sm
import re


def parse_variables_tz2():
    '''
    Функция возвращает кориеж состоящий из переменной, которая является sympy symbol
    :return:
    tuple of sympy symbol
    '''

    variables = input('Введите названия переменной ')

    return variables


def parse_function_tz2(variables):
    '''
    Запускает парсер функции. Возвращает sympy expression
    :param variables: tuple of sympy symbols
    :return:
    sympy expression
    '''

    function = input('Введите функцию: ')
    function = sm.parsing.sympy_parser.parse_expr(function.replace('–', '-'))
    function = sm.lambdify(variables, function)

    return function


def parse_restriction_tz2(variables):
    '''
    Функция возвращает словарь, где ключом выступают названия переменных,
    значениями являются списки, с ограничивающими интервалами
    :param variables:
    :return: dict: key=variables, value=list
    '''

    restr_values = input(f'Введите ограничения  ')
    restr_values = restr_values.strip()  # убираем пробелы в конце и начале

    restr_values = restr_values.replace(',', ' ')  # проверка вводов
    restr_values = re.sub(" +", " ", restr_values)  # убираем лишние пробелы
    restr_val = restr_values
    restr = restr_val.split()
    interval = [float(restr[0]), float(restr[1])]
    return interval


def parser_task_gr():
    '''
    Функция почередно проверяющая все входные данные
    :return: tuple containing of: tuple, sympy expression, dict
    '''
    variables = parse_variables_tz2()
    func = parse_function_tz2(variables)
    interval = parse_restriction_tz2(variables)
    return variables, func, interval


import sympy as sm
import numpy as np
import re

def parse_restriction(variables):
    '''
    Функция возвращает словарь, где ключом выступают названия переменных,
    значениями являются списки, с ограничивающими интервалами
    :param variables:
    :return: dict: key=variables, value=list
    '''
    restr_val = None
    flag_restriction = False
    while not flag_restriction:
        try:
            restrictions = input('Есть ли ограничения ? 1-да/ 0 – нет : ')
            if len(restrictions.strip()) == 1:
                restrictions = int(restrictions)
            else:
                raise Exception

            restr_val = {}
            if restrictions == 1:
                for var in variables:
                    restr_values = input(f'Введите ограничения по {str(var)}, если нет ограничений-введите 0:  -1 1 ')
                    restr_values = restr_values.strip()  # убираем пробелы в конце и начале
                    if restr_values == '0':
                        restr_val[var] = [-np.inf, np.inf]  # при отстутствии ограничений

                    else:
                        restr_values = restr_values.replace(',', ' ')  # проверка вводов
                        restr_values = re.sub(" +", " ", restr_values)  # убираем лишние пробелы

                        restr_val[var] = list(map(int, restr_values.split(' ')))
                        if len(restr_val[var]) != 2:
                            raise Exception
            elif restrictions == 0:
                for var in variables:
                    restr_val[var] = [-np.inf, np.inf]

            else:
                raise Exception

            flag_restriction = True
        except Exception:
            print('Ошибка ввода: ')
    return restr_val



import plotly.express as px
import numpy as np
import plotly.graph_objects as go

def plot(func, interval):
    """
       Функция отрисовки графиков функции двух переменных с отображением точек экстремума
       :param func: sympy expression
       :param interval: интервал функции
       :return: go.Figure содержаший два графика: поверхность функции и линии уровня
       """

    fig = px.scatter(df, x='point', y='function', size='size',
                     animation_frame='iter',
                     size_max=7,
                     title='График функции')
    x_ax = np.linspace(interval[0], interval[1], 200)
    y_ax = func(x_ax)
    fig.add_trace(go.Scatter(x=x_ax, y=y_ax, name='функция'))

    fig.update_layout(
        xaxis_title='значение х, у.е.',
        yaxis_title='значении f(x) у.е.')
    return fig


import sympy as sm
import numpy as np
import re

def parse__restr_function(variables):
    '''
    Функция возвращает словарь, где ключом выступают названия переменных,
    значениями являются списки, с ограничивающими интервалами
    :param variables:
    :return: dict: key=variables, value=list
    '''
    restr_function = None
    flag_func = False
    while not flag_func:

        try:
            restr_function = input('Введите  огранеичивающую функцию: ')
            restr_function = sm.parsing.sympy_parser.parse_expr(restr_function.replace('–', '-'))
#             check_var = restr_function.free_symbols.difference(variables)
#             if len(check_var) != 0:
#                 print('Введены лишние переменные')
#                 raise Exception
            flag_func = True
        except Exception:
            print('Ошибка ввода. Введите функцию еще раз: ')
    return restr_function


def gradient_descent_fixed(w_init, obj_func, grad_func, learning_rate=0.05,
                           max_iterations=500, threshold=1e-2):
    """
    Функция поиска минимума целевой функции одной переменной методом градиентного спуска, который находит оптимальное решение, делая шаг в сторону максимальной скорости уменьшения функции.
    :param obj_fun: objective function that is going to be minimized (call val,grad = obj_fun(w)).
    :param w_init: starting point Mx1
    :param grad_func: gradiation function that is going to be minimized (call val,grad = obj_fun(w)).
    :param max_iterations: number of epochs / iterations of an algorithm
    :param learning_rate: learning rate
    :return: function optimizes obj_fun using gradient descent. It returns (w,func_values),
    where w is optimal value of w and func_valus is vector of values of objective function [epochs x 1] calculated for each epoch
        """

    w = w_init
    w_history = w
    f_history = obj_func(w)
    delta_w = np.zeros(w.shape)
    i = 0
    diff = 1.0e10

    while i < max_iterations and diff > threshold:
        delta_w = -learning_rate * grad_func(w)
        w = w + delta_w

        # store the history of w and f
        w_history = np.vstack((w_history, w))
        f_history = np.vstack((f_history, obj_func(w)))

        # update iteration number and diff between successive values
        # of objective function
        i += 1
        diff = np.absolute(f_history[-1] - f_history[-2])
        print(f'X = {w_history[-1]}', f'Y = {f_history[-1]}', f'Iter = {i}', sep='\n')
        print({'Значение': f_history[-1], 'Кол-во итераций': i})
    return w_history, f_history


import pandas as pd


def parabolic_interpolation(func, interval, pression=1e-5, max_iter=500,
                            flag_results=False, flag_data=False, type_opt='min'):
    """
    Функция поиска экстремума функции методом парабол
    :param func: python function
    :param interval: list consisting of float numbers(left and right bounds)
    :param pression: optimization accuracy
    :param max_iter: maximum amount of iterations
    :param flag_results: output of intermediate results
    :param flag_data: recording intermediate results in a dataset
    :param type_opt: type of extrema: if min:flag_opt=1
                                      if max: flag_opt=-1
    :return: extrema:the point at which the extremum
             df: intermediate results in a dataset
    """

    if type_opt == 'min':
        flag_opt = 1
    else:
        flag_opt = -1

    data = []
    df = pd.DataFrame(columns=['iter', 'point', 'function'])
    x0 = interval[0]
    x1 = interval[1]
    x2 = (x0 + x1) / 2

    f0 = flag_opt * func(x0)
    f1 = flag_opt * func(x1)
    f2 = flag_opt * func(x2)
    f_x = {x0: f0, x1: f1, x2: f2}
    x2, x1, x0 = sorted([x0, x1, x2], key=lambda x: f_x[x])

    if flag_data:
        data = [[0, x2, func(x2)]]
    if flag_results:
        print(f"Номер итерации {0}, точка {x2}, функция {func(x2)} .")


    for i in range(max_iter):

        num = (x1 - x2) ** 2 * (f2 - f0) + (x0 - x2) ** 2 * (f1 - f2)
        den = 2 * ((x1 - x2) * (f2 - f0) + (x0 - x2) * (f1 - f2))
        x_extr = x2 + num / den

        if x_extr == x0 or x_extr == x1 or x_extr == x2:
            print('code 2, точка совпала с уже имеющимися')

        if den == 0:
            print('code 2, невозможно вычислить точку')

        if interval[0] > x_extr > interval[1]:
            print('code 2, точка вышла за интервал')

        f_extr = flag_opt * func(x_extr)

        if f_extr < f2:
            x0, f0 = x1, f1
            x1, f1 = x2, f2
            x2, f2 = x_extr, f_extr

        elif f_extr < f1:
            x0, f0 = x1, f1
            x1, f1 = x_extr, f_extr

        elif f_extr < f0:
            x0, f0 = x_extr, f_extr
        if i == 499:
            print('достигнуто максимальное количество итераций code 1')

        elif abs(x1 - x2) < pression and abs(f1 - f2) < pression:
            print('найдено значение с заданной точностью, code 0')
            break

        if flag_data:
            data.append(([i + 1, x_extr, func(x_extr)]))
            df = pd.DataFrame(data, columns=['iter', 'point', 'function'])

        if flag_results:
            print(f"Номер итерации {i + 1}, точка {x_extr}, функция {func(x_extr)}.")


    return x_extr, df


from Anar import Lagrange
from Anar import brent
from Anar import golden_ratio
from Anar import parse_variables_tz2
from Anar import parabolic_interpolation
from Anar import plot
from Anar import gradient_descent_fixed
from Anar import local_extrema
from Anar import gradient_descent_fixed
from Anar import parse__restr_function
from Anar import parse_restriction
from Anar import parse_variables
from Anar import parser_task
from Anar import tz4
from Anar import plot
from Anar import plot_tz2





import numpy as np
import pandas as pd
import random
import matplotlib.pyplot as plt
from matplotlib.pylab import rcParams
rcParams['figure.figsize'] = 12, 10

#Define input array with angles from 60deg to 300deg converted to radians
x = np.array([i*np.pi/180 for i in range(60,300,4)])
np.random.seed(10)  #Setting seed for reproducibility
y = np.sin(x) + np.random.normal(0,0.15,len(x))
data = pd.DataFrame(np.column_stack([x,y]),columns=['x','y'])
plt.plot(data['x'],data['y'],'.')


def parser_task1():
    '''
    Функция почередно проверяющая все входные данные
    :return: tuple containing of: tuple, sympy expression, dict
    '''
    variables = parse_variables()
    func = parse_function(variables)
    restr = parse_restriction(variables)
    return variables, func, restr